#!/usr/bin/env python
#
# Copyright 2012 Red Hat, Inc
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import json
import select
import sys
import threading
import traceback


__all__ = {"Flags", "Format", "Handler", "Plugin", "ThreadingPlugin"}

ERROR_PARSE_ERROR = (-32700, "Parse error.")
ERROR_INVALID_REQUEST = (-32600, "Invalid Request.")
ERROR_METHOD_NOT_FOUND = (-32601, "Method not found.")
ERROR_INVALID_PARAMS = (-32602, "Invalid params.")
ERROR_INTERNAL_ERROR = (-32603, "Internal error.")

METHOD_PREFIX = "authhub::"

def error(error, id=None):
    error = {"code": error[0], "message": error[1]}
    error = {"jsonrpc": "2.0", "error": error, "id": id}
    return json.dumps(error)

class Flags(object):
    nextOTP = (1 << 1)
    combine = (1 << 2)
    collectPIN = (1 << 3)
    doNotCollectPIN = (1 << 4)
    mustEncryptNonce = (1 << 5)
    separatePINRequired = (1 << 6)
    checkDigit = (1 << 7)

class Format(object):
    decimal = 0
    hexadecimal = 1
    alphanumeric = 2
    binary = 3
    base64 = 4

class Handler(object):
    def getTokenInfo(self, params):
        return []

    def verifyRequest(self, params):
        return False

class Sanitizer(object):
    TI_OCTET_STRINGS = ("otp-challenge", "otp-tokenID")
    REQ_OCTET_STRINGS = ("nonce", "otp-value", "otp-challenge",
                         "otp-counter", "otp-tokenID")

    def __init__(self, handler):
        self.handler = handler

    def getTokenInfo(self, params):
        result = self.handler.getTokenInfo(params)
        for ti in result:
            assert("otp-vendor" in ti) # This field is required
            if "flags" not in ti:
                ti["flags"] = 0
            for k, v in ti.items():
                if k in Sanitizer.TI_OCTET_STRINGS:
                    ti[k] = base64.b64encode(v)
        return result

    def verifyRequest(self, params):
        req = params.get("request", {})
        for k, v in req.items():
            if k in Sanitizer.REQ_OCTET_STRINGS:
                req[k] = base64.b64decode(v)
        return self.handler.verifyRequest(params)

class Plugin(object):
    def __init__(self, handler, rfile=sys.stdin, wfile=sys.stdout):
        self.rfile = rfile
        self.wfile = wfile
        self.handler = Sanitizer(handler)

    def respond(self, response):
        self.wfile.write(response + "\n")
        self.wfile.flush()

    def dispatch(self, method, params, id):
        try:
            rsp = {"jsonrpc": "2.0", "result": method(params), "id": id}
            rsp = json.dumps(rsp)
        except:
            rsp = error(ERROR_INTERNAL_ERROR, id)

        if id is not None:
            self.respond(rsp)

    def run(self):
        if select.select([self.rfile], [], [self.rfile])[2]:
            sys.stderr("Socket closed, exiting...")
            return False

        try:
            line = self.rfile.readline().strip()
            req = json.loads(line)
        except ValueError:
            try:
                traceback.print_exc()
                self.wfile.write(error(ERROR_PARSE_ERROR) + "\n")
            finally:
                return False

        # Ignore responses
        err = req.get("error", None)
        res = req.get("result", None)
        if err or res:
            return True

        # Make sure we have a valid request
        method = req.get("method", None)
        jsonrpc = req.get("jsonrpc", None)
        if jsonrpc != "2.0" or method is None:
            traceback.print_exc()
            self.wfile.write(error(ERROR_INVALID_REQUEST) + "\n")
            return True

        id = req.get("id", None)
        params = req.get("params", None)

        try:
            if not method.startswith(METHOD_PREFIX):
                raise AttributeError()
            method = method[len(METHOD_PREFIX):]

            if method.startswith("_"):
                raise AttributeError()
            self.dispatch(getattr(self.handler, method), params, id)
        except AttributeError:
            self.wfile.write(error(ERROR_METHOD_NOT_FOUND, id) + "\n")

        return True

    def runForever(self):
        while self.run():
            continue

class ThreadingPlugin(Plugin):
    def __init__(self, *args, **kwargs):
        Plugin.__init__(self, *args, **kwargs)
        self.lock = threading.Lock()

    def respond(self, response):
        self.lock.acquire()
        Plugin.respond(self, response)
        self.lock.release()

    def dispatch(self, method, params, id):
        threading.Thread(target=Plugin.dispatch,
                         args=(self, method, params, id)).start()
